unit DSMandel;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls,
  Forms, Dialogs, ExtCtrls;

type
  TDSPos = record
    x: integer;
    y: integer;
  end;

  TDSConfigRecord = record
    MaxIter: integer;
    XMin: Extended;
    XMax: Extended;
    YMin: Extended;
    YMax: Extended;
  end;

  TDSColors256 = array[0..$FF] of TColor;

  TDSLine = array of Byte;

  TDSMandelMode = (mbIdle, mbRendering);
  TDSRefreshMode = (rmTimer, rmIncremental);
  TDSIncrementMode = (imLine, imBitmap);

  TDSMouseMoveEvent = procedure(Sender: TObject; Shift: TShiftState;
  X, Y: Integer) of object;
  TDSMouseButtonEvent = procedure(Sender: TObject;
    Button: TMouseButton; Shift: TShiftState; X, Y: Integer) of object;

  TDSMandel = class(TWinControl)
  private
    { Dclarations prives }
    FPaintBox: TPaintBox;
    FTimer: TTimer;
    FBitmap: TBitmap;
    FThread: TThread;
    FThreadPriority: TThreadPriority;
    FRefreshMode: TDSRefreshMode;
    FIncrementMode: TDSIncrementMode;
    FMode: TDSMandelMode;
    FColors: TDSColors256;
    FCursor: TCursor;
    FPos: TDSPos;
    FSelStart: TDSPos;
    FSelEnd: TDSPos;
    FLine: TDSLine;
    FXMin: Extended;
    FXMax: Extended;
    FYMin: Extended;
    FYMax: Extended;
    FMaxIter: Integer;
    FPercent: Integer;
    FWidth: Integer;
    FHeight: Integer;
    FSelecting: Boolean;
    j: integer;
    FOnFinish: TNotifyEvent;
    FOnPercentChange: TNotifyEvent;
    FOnReset: TNotifyEvent;
    FOnResize: TNotifyEvent;
    FOnStart: TNotifyEvent;
    FOnStop: TNotifyEvent;
    FOnChangeZoom: TNotifyEvent;
    FOnMouseMove: TDSMouseMoveEvent;
    FOnMouseDown: TDSMouseButtonEvent;
    FOnMouseUp: TDSMouseButtonEvent;
    procedure PaintBoxPaint(Sender: TObject);
    procedure PaintBoxMouseMove(Sender: TObject;
      Shift: TShiftState; X, Y: Integer);
    procedure PaintBoxMouseUp(Sender: TObject;
      Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure PaintBoxMouseDown(Sender: TObject;
      Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure TimerTimer(Sender: TObject);
    procedure SetXMin(Value: Extended);
    procedure SetXMax(Value: Extended);
    procedure SetYMin(Value: Extended);
    procedure SetYMax(Value: Extended);
    procedure SetMaxIter(Value: Integer);
    procedure SetWidth(Value: Integer);
    procedure SetHeight(Value: Integer);
    procedure SetPercent(Value: Integer);
    procedure SetPosX(Value: Integer);
    procedure SetPosY(Value: Integer);
    procedure SetThreadPriority(Value: TThreadPriority);
    procedure SetMode(Value: TDSMandelMode);
    procedure SetLine(x: Integer; Value: Byte);
    procedure SetInterval(Value: integer);
    procedure SetCursor(Value: TCursor);
    function GetInterval: integer;
    function GetLine(x: Integer): Byte;
    function GetPosX: Integer;
    function GetposY: Integer;
    function GetBitmap: TBitmap;
  protected
    { Dclarations protges }
  public
    { Dclarations publiques }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    property Bitmap: TBitmap read GetBitmap;
    property Percent: Integer read FPercent write SetPercent;
    property PosX: Integer read GetPosX write SetPosX;
    property PosY: Integer read GetPosY write SetPosY;
    property Line[x: integer]: Byte read GetLine write SetLine;
    procedure SetBounds(ALeft, ATop, AWidth, AHeight: Integer); override;
    procedure Reset;
    procedure Start;
    procedure Stop;
    procedure DrawLine;
    procedure Finish;
    procedure SaveBitmapToFile(const FileName: string);
    procedure SaveParamsToFile(const FileName: string);
    procedure ReadParamsFromFile(const FileName: string);
  published
    { Dclarations publies }
    property OnClick;
    property OnDblClick;
    property OnMouseDown: TDSMouseButtonEvent read FOnMouseDown write FOnMouseDown;
    property OnMouseMove: TDSMouseMoveEvent read FOnMouseMove write FOnMouseMove;
    property OnMouseUp: TDSMouseButtonEvent read FOnMouseUp write FOnMouseUp;
    property OnChangeZoom: TNotifyEvent read FOnChangeZoom write FOnChangeZoom;
    property OnFinish: TNotifyEvent read FOnFinish write FOnFinish;
    property OnPercentChange: TNotifyEvent read FOnPercentChange write FOnPercentChange;
    property OnReset: TNotifyEvent read FOnReset write FOnReset;
    property OnResize: TNotifyEvent read FOnResize write FOnResize;
    property OnStart: TNotifyEvent read FOnStart write FOnStart;
    property OnStop: TNotifyEvent read FOnStop write FOnStop;
    property XMin: Extended read FXMin write SetXMin;
    property XMax: Extended read FXMax write SetXMax;
    property YMin: Extended read FYMin write SetYMin;
    property YMax: Extended read FYMax write SetYMax;
    property MaxIter: Integer read FMaxIter write SetMaxIter;
    property Width: Integer read FWidth write SetWidth;
    property Height: Integer read FHeight write SetHeight;
    property ThreadPriority: TThreadPriority read FThreadPriority write SetThreadPriority;
    property Mode: TDSMandelMode read FMode write SetMode;
    property RefreshInterval: Integer read GetInterval write SetInterval;
    property RefreshMode: TDSRefreshMode read FRefreshMode write FRefreshMode;
    property IncrementMode: TDSIncrementMode read FIncrementMode write FIncrementMode;
    property Cursor: TCursor read FCursor write SetCursor;
    property Align;
    property Visible;
    property ShowHint;
    property Parent;
    property ParentShowHint;
  end;

  TDSMandelThread = class(TThread)
  private
    { Dclarations prives }
    FOwner: TDSMandel;
    FPos: TDSPos;
    FHeight: integer;
    FWidth: integer;
    FXMin: extended;
    FXMax: extended;
    FYMin: extended;
    FYMax: extended;
    FMaxIter: integer;
    x, y, x2, y2: extended;
    temp: extended;
    cx, cy: extended;
    level:integer;
    iter: integer;
    ctr: integer;
  public
    { Dclarations publiques }
    constructor Create(AOwner: TDSMandel);
    procedure execute; override;
  published
    { Dclarations publies }
    property Height: integer read FHeight write FHeight;
    property Width: integer read FWidth write FWidth;
    property XMin: extended read FXMin write FXMin;
    property XMax: extended read FXMax write FXMax;
    property YMin: extended read FYMin write FYMin;
    property YMax: extended read FYMax write FYMax;
    property MaxIter: integer read FMaxIter write FMaxIter;
    property Position: TDSPos read FPos write FPos;
  end;

function min(a, b: integer): integer;
function max(a, b: integer): integer;

procedure Register;

implementation

constructor TDSMandel.Create(AOwner: TComponent);
var
  i: Byte;
  pal: PLogPalette;
  hpal: HPALETTE;
begin
  inherited Create(AOwner);
  FCursor := crDefault;
  FRefreshMode := rmIncremental;
  FIncrementMode := imLine;
  FSelecting := False;
  FMaxIter := 200;
  FXMin := -3;
  FXMax := 2;
  FYMin := -2;
  FYMax := 2;
  FMode := mbIdle;
  FBitmap := TBitmap.Create;
  FBitmap.PixelFormat := pf8bit;
  for i := 0 to 255 do
    begin
      FColors[i] := rgb(i, i, 55 + i * 200 div 255);
    end;
  pal := nil;
  try
    GetMem(Pal, SizeOf(TLogPalette) + SizeOf(TPaletteEntry) * 255);
    pal.palVersion := $300;
    pal.palNumEntries := 256;
    for i := 0 to 255 do
      begin
        pal.palPalEntry[i].peRed := i;
        pal.palPalEntry[i].peGreen := i;
        pal.palPalEntry[i].peBlue := 55 + i * 200 div 255;
      end;
    hpal := CreatePalette(pal^);
    if hpal <> 0 then FBitmap.Palette := hpal;
  finally
    Freemem(pal);
  end;
  FThreadPriority := tpNormal;
  FPaintBox := TPaintBox.Create(Self);
  FPaintBox.Parent := Self;
  FPaintBox.OnPaint := PaintBoxPaint;
  FPaintBox.OnMouseDown := PaintBoxMouseDown;
  FPaintBox.OnMouseMove := PaintBoxMouseMove;
  FpaintBox.OnMouseUp := PaintBoxMouseUp;
  FTimer := TTimer.Create(Self);
  FTimer.Enabled := False;
  FTimer.Interval := 50;
  FTimer.OnTimer := TimerTimer;
  FThread := TDSMandelThread.Create(Self);
  SetBounds(Left, Top, 320, 240);
end;

destructor TDSMandel.Destroy;
begin
  FTimer.Free;
  FThread.Suspend;
  FThread.Free;
  FBitmap.Free;
  FPaintBox.Free;
  inherited Destroy;
end;

procedure TDSMandel.SetBounds(ALeft, ATop, AWidth, AHeight: Integer);
begin
  inherited SetBounds(ALeft, ATop, AWidth, AHeight);
  FWidth := AWidth;
  FHeight := AHeight;
  FPaintBox.Width := AWidth;
  FPaintBox.Height := AHeight;
  SetLength(FLine, AWidth);
  if Assigned(FBitmap) then
    begin
      FBitmap.Width := AWidth;
      FBitmap.Height := AHeight;
    end;
  Reset;
  if Assigned(FOnResize) then FOnResize(Self);
end;

procedure TDSMandel.PaintBoxPaint(Sender: TObject);
begin
  if Assigned(FBitmap) then FPaintBox.Canvas.Draw(0, 0, FBitmap);
  if csDesigning in ComponentState then
    begin
      FPaintBox.Canvas.Pen.Style := psDash;
      FPaintBox.Canvas.Brush.Style := bsClear;
      FPaintBox.Canvas.Rectangle(0, 0, FWidth, FHeight);
    end;
end;

procedure TDSMandel.SetXMin(Value: Extended);
begin
  FXMin := Value;
  Reset;
  if Assigned(FOnChangeZoom) then FOnChangeZoom(Self);
end;

procedure TDSMandel.SetXMax(Value: Extended);
begin
  FXMax := Value;
  Reset;
  if Assigned(FOnChangeZoom) then FOnChangeZoom(Self);
end;

procedure TDSMandel.SetYMin(Value: Extended);
begin
  FYMin := Value;
  Reset;
  if Assigned(FOnChangeZoom) then FOnChangeZoom(Self);
end;

procedure TDSMandel.SetYMax(Value: Extended);
begin
  FYMax := Value;
  Reset;
  if Assigned(FOnChangeZoom) then FOnChangeZoom(Self);
end;

function TDSMandel.GetBitmap: TBitmap;
begin
  if FBitmap = nil then
    begin
      FBitmap := TBitmap.Create;
      FBitmap.Width := FWidth;
      FBitmap.Height := FHeight;
    end;
  Result := FBitmap;
end;

procedure TDSMandel.SetMaxIter(Value: Integer);
begin
  FMaxIter := Value;
  Reset;
  if Assigned(FOnChangeZoom) then FOnChangeZoom(Self);
end;

procedure TDSMandel.SetWidth(Value: Integer);
begin
  FWidth := Value;
  SetBounds(Left, Top, Value, FHeight);
end;

procedure TDSMandel.SetHeight(Value: Integer);
begin
  FHeight := Value;
  SetBounds(Left, Top, FWidth, Value);
end;

procedure TDSMandel.SetPercent(Value: Integer);
begin
  FPercent := Value;
  if Assigned(FOnPercentChange) then FOnPercentChange(Self);
end;

procedure TDSMandel.Reset;
begin
  stop;
  FPos.x := 0;
  FPos.y := 0;
  Percent := 0;
  if Assigned(FOnReset) then FOnReset(Self);
end;

procedure TDSMandel.SetPosX(Value: Integer);
begin
  FPos.x := Value;
end;

procedure TDSMandel.SetPosY(Value: Integer);
begin
  FPos.y := Value;
end;

function TDSMandel.GetPosX: Integer;
begin
  Result := FPos.x;
end;

function TDSMandel.GetPosY: Integer;
begin
  Result := FPos.y;
end;

procedure TDSMandel.SetThreadPriority(Value: TThreadPriority);
begin
  FThreadPriority := Value;
  FThread.Priority := Value;
end;

procedure TDSMandel.SetMode(Value: TDSMandelMode);
begin
  FMode := Value;
  if Value = mbRendering then Start
  else Stop;
  if (Value = mbRendering) and (Assigned(FOnStart)) then FOnStart(Self);
  if (Value = mbIdle) and (Assigned(FOnStop)) then FOnStop(Self);
end;

procedure TDSMandel.DrawLine;
var
  i: integer;
  p: pByteArray;
begin
  j := PosY;
  p := FBitmap.ScanLine[j];
  for i := 0 to Pred(FWidth) do
    begin
      p[i] := FLine[i];
      if (FRefreshMode = rmIncremental) and (FIncrementMode = imLine) then
        FPaintBox.Canvas.Pixels[i, j] := FColors[FLine[i]];
    end;
  if (FRefreshMode = rmIncremental) and (FIncrementMode = imBitmap) then
        FPaintBox.Canvas.Draw(0, 0, FBitmap);
end;

procedure TDSMandel.Finish;
begin
  if Assigned(FOnFinish) then FOnFinish(Self);
  Reset;
end;

procedure TDSMandel.Start;
begin
  if FRefreshMode = rmTimer then FTimer.Enabled := True;
  repeat
    FThread.Resume;
  until not FThread.Suspended;
  FMode := mbRendering;
  if Assigned(FOnStart) then FOnStart(Self);
end;

procedure TDSMandel.Stop;
begin
  FPaintbox.Invalidate;
  FTimer.Enabled := False;
  FThread.Suspend;
  FMode := mbIdle;
  if Assigned(FOnStop) then FOnStop(Self);
end;

procedure TDSMandel.SetLine(x: Integer; Value: Byte);
begin
  Fline[x] := Value;
end;

function TDSMandel.GetLine(x: Integer): Byte;
begin
  Result := FLine[x];
end;

procedure TDSMandel.SetCursor(Value: TCursor);
begin
  if FCursor <> Value then
  begin
    FCursor := Value;
    FPaintBox.Cursor := Value;
    Perform(CM_CURSORCHANGED, 0, 0);
  end;
end;

procedure TDSMandel.SetInterval(Value: integer);
begin
  FTimer.Interval := Value;
end;

function TDSMandel.GetInterval: integer;
begin
  Result := FTimer.Interval;
end;

procedure TDSMandel.PaintBoxMouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
var
  a, b, c, d, e, f: integer;
begin
  if (FSelecting = False) and (ssLeft in Shift) then
    begin
      FSelecting := True;
      FSelStart.x := x;
      FSelStart.y := y;
    end
  else
  if (ssLeft in Shift) then
    begin
      FPaintBox.Canvas.Draw(0, 0, FBitmap);
      FSelEnd.x := x;
      FSelEnd.y := y;
      if (abs(x - FSelStart.x) > abs(y - FSelStart.y)*FWidth div FHeight) then
        begin
          if FSelEnd.y > FSelStart.y then
            begin
              FSelEnd.y := FSelStart.y + abs(FSelEnd.x - FSelStart.x) * FHeight div FWidth;
            end
          else
            begin
              FSelEnd.y := FSelStart.y - abs(FSelEnd.x - FSelStart.x) * FHeight div FWidth;
            end;
        end
      else
        begin
          if FSelEnd.x > FSelStart.x then
            begin
              FSelEnd.x := FSelStart.x + abs(FSelEnd.y - FSelStart.y) * FWidth div FHeight;
            end
          else
            begin
              FSelEnd.x := FSelStart.x - abs(FSelEnd.y - FSelStart.y) * FWidth div FHeight;
            end;
        end;
        a := FSelStart.x;
        b := FSelStart.y;
        c := FSelEnd.x;
        d := FSelEnd.y;
        e := a + a - c;
        f := b + b - d;
        FPaintBox.Canvas.Pixels[a, b] := clDkGray;
        FPaintBox.Canvas.DrawFocusRect(Rect(min(c, e), min(d, f), max(c, e), max(d, f)));
    end
  else
    begin
      FSelecting := False;
    end;
  if Assigned(FOnMouseMove) then FOnMouseMove(Sender, Shift, X, Y);
end;

procedure TDSMandel.PaintBoxMouseDown(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  TXMin: Extended;
  TXMax: Extended;
  TYMin: Extended;
  TYMax: Extended;
  a, b, c, d: Integer;
begin
  if Button = mbRight then
    begin
      FBitmap.Canvas.Rectangle(0, 0, FWidth+1, FHeight+1);
      FPaintBox.Canvas.Draw(0, 0, FBitmap);
      FSelecting := False;
      a := X - FWidth;
      b := X + FWidth;
      c := Y - FHeight;
      d := Y + FHeight;
      TXMin := (FXMax - FXMin) / FWidth * min(a, b) + FXMin;
      TXMax := (FXMax - FXMin) / FWidth * max(a, b) + FXMin;
      TYMin := (FYMax - FYMin) / FHeight * min(c, d) + FYMin;
      TYMax := (FYMax - FYMin) / FHeight * max(c, d) + FYMin;
      FXMin := TXMin;
      FXMax := TXMax;
      FYMin := TYMin;
      FYMax := TYMax;
      if Assigned(FOnChangeZoom) then FOnChangeZoom(Self);
      Reset;
      Start;
    end;
  if Assigned(FOnMouseDown) then FOnMouseDown(Sender, Button, Shift, X, Y);
end;

procedure TDSMandel.PaintBoxMouseUp(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  TXMin: Extended;
  TXMax: Extended;
  TYMin: Extended;
  TYMax: Extended;
  a,b: integer;
begin
  if Button = mbLeft then
    begin
      FSelecting := False;
      a := FSelStart.x + FSelStart.x - FSelEnd.x;
      b := FSelStart.y + FSelStart.y - FSelEnd.y;
      FBitmap.Canvas.CopyRect(Rect(0, 0, FWidth, FHeight), FBitmap.Canvas, Rect(min(a, FSelEnd.x), min(b, FSelEnd.y), max(a, FSelEnd.x), max(b, FSelEnd.y)));
      FPaintBox.Canvas.Draw(0, 0, FBitmap);
      TXMin := (FXMax - FXMin) / FWidth * min(FSelEnd.x, a) + FXMin;
      TXMax := (FXMax - FXMin) / FWidth * max(FSelEnd.x, a) + FXMin;
      TYMin := (FYMax - FYMin) / FHeight * min(FSelEnd.y, b) + FYMin;
      TYMax := (FYMax - FYMin) / FHeight * max(FSelEnd.y, b) + FYMin;
      FXMin := TXMin;
      FXMax := TXMax;
      FYMin := TYMin;
      FYMax := TYMax;
      if Assigned(FOnChangeZoom) then FOnChangeZoom(Self);
      Reset;
      Start;
    end;
  if Assigned(FOnMouseUp) then FOnMouseUp(Sender, Button, Shift, X, Y);
end;

procedure TDSMandel.TimerTimer(Sender: TObject);
var
  a, b, c, d, e ,f : integer;
begin
  FPaintBox.Canvas.Draw(0, 0, FBitmap);
  if FSelecting then
    begin
      a := FSelStart.x;
      b := FSelStart.y;
      c := FSelEnd.x;
      d := FSelEnd.y;
      e := a + a - c;
      f := b + b - d;
      FPaintBox.Canvas.Pixels[a, b] := clDkGray;
      FPaintBox.Canvas.DrawFocusRect(Rect(min(c, e), min(d, f), max(c, e), max(d, f)));
    end;
  if FRefreshMode = rmIncremental then FTimer.Enabled := False;
end;

procedure TDSMandel.SaveBitmapToFile(const FileName: string);
begin
  FBitmap.SaveToFile(FileName);
end;

procedure TDSMandel.SaveParamsToFile(const FileName: string);
var
  Stream: File of TDSConfigRecord;
  Config: TDSConfigRecord;
begin
  AssignFile(Stream, FileName);
  Config.MaxIter := FMaxIter;
  Config.XMin := FXMin;
  Config.XMax := FXMax;
  Config.YMin := FYMin;
  Config.YMax := FYMax;
  try
    FileMode := 1;
    Rewrite(Stream);
    Write(Stream, Config);
  finally
    Closefile(Stream);
  end;
end;

procedure TDSMandel.ReadParamsFromFile(const FileName: string);
var
  Stream: File of TDSConfigRecord;
  Config: TDSConfigRecord;
begin
  if not fileexists(FileName) then
    begin
      MessageDlg('Le Fichier spcifi n''existe pas', mtError, [mbOk], 0);
      Exit;
    end;
  AssignFile(Stream, FileName);
  try
    FileMode := 0;
    System.Reset(Stream);
    Read(Stream, Config);
    XMin := Config.XMin;
    XMax := Config.XMax;
    YMin := Config.YMin;
    YMax := Config.YMax;
    MaxIter := Config.MaxIter;
    Reset;
  finally
    CloseFile(Stream);
  end;
end;

constructor TDSMandelThread.Create(AOwner: TDSMandel);
begin
  inherited Create(True);
  FOwner := AOwner;
  Priority := FOwner.ThreadPriority;
  FWidth := FOwner.Width;
  FHeight := FOwner.Height;
  FXMin := FOwner.XMin;
  FXMax := FOwner.XMax;
  FYMin := FOwner.YMin;
  FYMax := FOwner.YMax;
  FMaxIter := FOwner.MaxIter;
end;

procedure TDSMandelThread.Execute;
begin
  repeat
    if FOwner.Mode = mbRendering then
      begin
        FPos.x := FOwner.PosX;
        FPos.y := FOwner.PosY;
        FXMin := FOwner.XMin;
        FYMin := FOwner.YMin;
        FXMax := FOwner.XMax;
        FYMax := FOwner.YMax;
        FMaxIter := FOwner.MaxIter;
        FWidth := FOwner.Width;
        FHeight := FOwner.Height;
        if FPos.Y > FHeight then
          begin
            Synchronize(FOwner.Reset);
          end;
        cy := FYMin + FPos.y*(FYMax-FYMin)/(FHeight-1);
        cx := FXmin + FPos.x*(FXMax-FXMin)/(FWidth-1);
        x := 0;
        y := 0;
        x2 := 0;
        y2 := 0;
        iter := 0;
        while ((iter < FMaxiter) and (x2+y2 < 10000)) do
          begin
            temp := x2 - y2 + cx;
            y := 2 * x * y + cy;
            x := temp;
            x2 := x * x;
            y2 := y * y;
            iter := iter + 1;
          end;
        level := iter + 48;
        if level < 255 then FOwner.Line[Fpos.x] := level
        else FOwner.Line[Fpos.x] := 255;
        ctr := 100 * Round(FPos.y*FWidth+FPos.x) div (FWidth * FHeight);
        if FOwner.Percent <> ctr then FOwner.Percent := ctr;
        FPos.X := FPos.X + 1;
        if FPos.x = FWidth then
          begin
            Synchronize(FOwner.DrawLine);
            FPos.X := 0;
            FPos.Y := FPos.Y + 1;
          end;
        if (FOwner.Mode = mbRendering) and ((FPos.Y = FOwner.PosY) or (FPos.Y = FOwner.PosY + 1)) then
          begin
            FOwner.PosX := FPos.X;
            FOwner.PosY := FPos.Y;
          end;
        if FPos.Y = FHeight then
          begin
            Synchronize(FOwner.Finish);
          end;
      end;
  until Terminated;
end;

function min(a, b: integer): integer;
begin
  if a < b then Result := a
  else Result := b;
end;

function max(a, b: integer): integer;
begin
  if a > b then Result := a
  else Result := b;
end;

procedure Register;
begin
  RegisterComponents('Dark Skull', [TDSMandel]);
end;

end.
